/*
 * Main entrypoint for Codecogs Equation Editor plugin
 * Use this plugin in your main CKEditor JS file like:
 * 
 * import Eqneditor5 from 'PATH/TO/DIR/eqneditor5';
 * 
 * ClassicEditor
 *	.create( document.querySelector( '#editor' ), {
 *		plugins: [
 *			Eqneditor5,
 *			...
 * 		],
 * 		toolbar: [
 * 			'eqneditor5Button',
 * 			...
 * 		]
 * 		...
 */

import { Plugin } from '@ckeditor/ckeditor5-core/src';
import { ButtonView, ContextualBalloon, clickOutsideHandler } from '@ckeditor/ckeditor5-ui/src';
import DomEventObserver from '@ckeditor/ckeditor5-engine/src/view/observer/domeventobserver';
import { default as EqnEditorView } from './dialog';
import fxIcon from './fx.svg';

class DoubleClickObserver extends DomEventObserver {
    constructor( view ) {
        super( view );

        this.domEventType = 'dblclick';
    }

    onDomEvent( domEvent ) {
        this.fire( domEvent.type, domEvent );
    }
}

export default class Eqneditor5 extends Plugin {
	static get pluginName() {
		return 'Equation Editor 5';
	}

	static get requires() {
        return [ ContextualBalloon ];
    }


	init() {
		const editor = this.editor;

		var host='latex.codecogs.com';
	  	var http=('https:' == window.location.protocol ? 'https://' : 'http://');

		this._installExtScript(http+host+'/js/eqneditor.api.min.js');
		this._installExtCSS(http+host+'/css/eqneditor_1.css');

        // Create the balloon and the form view.
        // This does not at this point contain the editor, as the scripts
        // for the editor will loaded in the background, triggered by the lines above.
        this._balloon = this.editor.plugins.get( ContextualBalloon );
        this.eqnView = this._createEqnEditorView();
        this.editorReady = false;

		// Add the "eqneditor5" button to feature components.
		editor.ui.componentFactory.add( 'EqnEditor5', locale => {
			const button = new ButtonView( locale );

			button.set( {
				label: editor.t( 'Equation Editor 5' ),
				icon: fxIcon,
				tooltip: true
			} );


			// Insert a text into the editor after clicking the button.
			button.on('execute', () => {
				this._initEqnEditor(locale);
			} );

			return button;
		} );
	
		// Listen for double click
		const view = editor.editing.view;
		const viewDocument = view.document;

		view.addObserver( DoubleClickObserver );
		editor.listenTo( viewDocument, 'dblclick', function(e, dat) {
			if (dat.domTarget.nodeName.toLowerCase() == "img") {
				var sName = dat.domTarget.src.match( /(gif|svg)\.image\?(.*)/ );
				if(sName!=null) this._initEqnEditor(null, sName[2]);
			}
		}.bind(this) );
	}

	_installExtScript( url ) {
		var scriptTag = document.createElement('script');
		scriptTag.src = url;
		scriptTag.type = 'text/javascript';
		document.head.appendChild(scriptTag);
	}

	_installExtCSS( url ) {
		var link = document.createElement('link');
		link.rel = "stylesheet";
		link.href = url;
		document.head.appendChild(link);
	}

	_initEqnEditor( locale, inp="" ) {
		this._showUI( locale );

		if(!this.editorReady)
		{
			this.output = new EqEditor.Output('output');
			this.textarea = EqEditor.TextArea.link('latexInput')
				.addOutput(this.output)
				.addHistoryMenu(new EqEditor.History('history'));
				
			EqEditor.Toolbar.link('toolbar').addTextArea(this.textarea);

			var nodes = document.getElementById('history').childNodes;
			for(var i=0; i<nodes.length; i++) { nodes[i].style.padding = 'revert'; }

			this.editorReady=true; // don't need to do this section again.
		}

		this.textarea.clear();
        this.textarea.insert(inp.replace(/&space;/g, ' '), inp.length);
	}

	_createEqnEditorView() {
        const editor = this.editor;
        const view = new EqnEditorView( editor.locale );

		this.listenTo( view, 'submit', () => {
			const viewFragment = this.editor.data.processor.toView( this.output.exportAs('html') );
			const modelFragment = this.editor.data.toModel( viewFragment );
			this.editor.model.insertContent(modelFragment);

			this.textarea.pushToHistory();

            // Hide the form view after submit.
            this._hideUI();
        } );

        // Hide the form view after clicking the "Cancel" button.
        this.listenTo( view, 'cancel', () => {
            this._hideUI();
        } );

        // Hide the form view when clicking outside the balloon.
        clickOutsideHandler( {
            emitter: view,
            activator: () => this._balloon.visibleView === view,
            contextElements: [ this._balloon.view.element ],
            callback: () => this._hideUI()
        } );

        return view;
    }

	_hideUI() {
		// Hack to avoid repeated adding when editor is clicked multiple times
		/*document.getElementById('history').innerHTML = '';
		document.getElementById('toolbar').innerHTML = '';*/
		document.getElementById('output').src = '';

		this.textarea.clear();

        this.eqnView.element.reset();

        this._balloon.remove( this.eqnView );

        // Focus the editing view after closing the form view.
        this.editor.editing.view.focus();
    }

	_isElementEditable(element) {
	    // Check if the element is a valid HTML element
	    if (!(element instanceof HTMLElement)) {
	        return false;
	    }

	    // Check if the element is part of the CKEditor editable area
	    return element.closest('.ck-editor__editable') !== null;
	}


	_findClosestEditableParent(element) {
	    while (element && !this._isElementEditable(element)) {
	        element = element.parentElement;
	    }
	    return element;
	}

	_getBalloonPositionData() {
        const view = this.editor.editing.view;
        const viewDocument = view.document;

        // Get the first range from the current selection
	    const firstRange = viewDocument.selection.getFirstRange();

	    if (!firstRange) {
	        // If no valid range is found, provide a default or handle the error
	        console.warn('No valid selection range found for balloon positioning.');
	        return {
	            target: null // Or provide a fallback position if necessary
	        };
	    }

	    // Get the DOM target for the balloon based on the range
	    let domTarget = view.domConverter.viewRangeToDom(firstRange);

	    // Check if the target is valid and if it's editable
	    if (!domTarget || !domTarget.getBoundingClientRect || !this._isElementEditable(domTarget)) {
	        // If target is non-editable or invalid, find the nearest editable parent or fallback to editor
	        domTarget = this._findClosestEditableParent(domTarget) || document.querySelector('.ck-content');
	    }

	    return {
	        target: () => domTarget
	    };
    }

	_showUI( locale ) {
		//const balloon = this.editor.plugins.get( ContextualBalloon );

	    // Check if the balloon is already visible to avoid repositioning it unnecessarily
	    if ( this._balloon.visibleView === this.eqnView ) {
	        return;
	    }

	    // Remove the current view if a different view is visible
	    if (this._balloon.visibleView) {
	        this._balloon.remove(balloon.visibleView);
	    }

		this._balloon.add( {
            view: this.eqnView,
            singleViewMode:true,
            position: this._getBalloonPositionData()
        } );

		// Force the balloon to update its position immediately
		this._balloon.showStack('main');
    	this._balloon.updatePosition( this._getBalloonPositionData() );

		// Causes the ballon tab to appear in the right place
		setTimeout(() => {
			window.dispatchEvent(new Event('resize'));
		}, 0);
    }
}
